@@ -1,3 +1,25 @@
+/* Copyright 2022-2024 Jaakko Keränen <jaakko.keranen@iki.fi>
+
+Redistribution and use in source and binary forms, with or without
+modification, are permitted provided that the following conditions are met:
+
+1. Redistributions of source code must retain the above copyright notice, this
+list of conditions and the following disclaimer.
+2. Redistributions in binary form must reproduce the above copyright notice,
+this list of conditions and the following disclaimer in the documentation
+and/or other materials provided with the distribution.
+
+THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
+ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
+ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */
+
 #include "render.h"
 #include "canvas.h"
 #include "video.h"
@@ -16,7 +38,7 @@ void SDL_ShowCursor(void); /* keyboard.c */
 #define FIRST_COLOR_PAIR    1
 
 iDeclareType(ColorPair)
-    
+
 struct Impl_ColorPair {
     uint8_t col[2];
 };
@@ -48,7 +70,7 @@ static void SDL_ResetColors(SDL_Renderer *d) {
 }
 
 iDeclareType(HSLColor)
-    
+
 struct Impl_HSLColor {
     float hue;
     float sat;
@@ -164,7 +186,7 @@ static SDL_ColorId SDL_GetColorId(SDL_Renderer *d, SDL_Color rgb) {
                 diff[0] -= 360;
             }
             else if (diff[0] < -180) {
-                diff[0] += 360; 
+                diff[0] += 360;
             }
             diff[0] = iAbs(diff[0]) / 180.0f;
 //            float delta;
@@ -384,9 +406,10 @@ void SDL_RenderClear(SDL_Renderer *d) {
     clear_Canvas(activeCanvas_(d), d->drawColorId);
 }
 
+
 void SDL_RenderPresent(SDL_Renderer *d) {
     curs_set(0);
-    present_Canvas(&d->defaultTarget->canvas, d->window->cwin);   
+    present_Canvas(&d->defaultTarget->canvas, d->window->cwin);
     wrefresh(d->window->cwin);
     if (SDL_IsTextInputStarted()) {
         SDL_ShowCursor();
@@ -407,7 +430,7 @@ void SDL_RenderCopy(SDL_Renderer *d, SDL_Texture *texture, const SDL_Rect *src,
     }
     blit_Canvas(activeCanvas_(d),
                 &texture->canvas,
-                *(const iRect *) src,
+                init_Rect(src->x, src->y, src->w, src->h),
                 dst ? init_I2(dst->x, dst->y) : zero_I2(),
                 texture->blend);
 }
@@ -417,7 +440,7 @@ int drawColor_(SDL_Renderer *d) {
 }
 
 void SDL_RenderDrawLines(SDL_Renderer *d, const SDL_Point *points, uint32_t count) {
-    /* This function is expected to be used only in a limited manner, to draw full or 
+    /* This function is expected to be used only in a limited manner, to draw full or
        half rectangles. */
     iCanvas  *canvas      = activeCanvas_(d);
     const int drawColorId = drawColor_(d);
@@ -432,6 +455,12 @@ void SDL_RenderDrawLines(SDL_Renderer *d, const SDL_Point *points, uint32_t coun
             setChar_Canvas(canvas, init_I2(points[0].x, points[0].y), '[', d->drawColorId);
             setChar_Canvas(canvas, init_I2(points[1].x - 1, points[1].y), ']', d->drawColorId);
         }
+        else if (rect.size.y == 2) {
+            setChar_Canvas(canvas, init_I2(points[0].x, points[0].y), 0x250c, d->drawColorId);
+            setChar_Canvas(canvas, init_I2(points[1].x - 1, points[1].y), 0x2510, d->drawColorId);
+            setChar_Canvas(canvas, init_I2(points[0].x, points[3].y - 1), 0x2514, d->drawColorId);
+            setChar_Canvas(canvas, init_I2(points[1].x - 1, points[2].y - 1), 0x2518, d->drawColorId);
+        }
     }
     else if (count == 3) {
         setChar_Canvas(canvas, init_I2(points[0].x, points[0].y), '|', d->drawColorId);
@@ -450,7 +479,7 @@ void SDL_RenderDrawLines(SDL_Renderer *d, const SDL_Point *points, uint32_t coun
             (iRect){ init_I2(points[0].x, points[0].y), init_I2(1, points[1].y - points[0].y) },
             transparent_VisChar,
             0x2502 /* box drawings light vertical */,
-            drawColorId);        
+            drawColorId);
     }
 }
 
@@ -464,7 +493,7 @@ void SDL_RenderFillRect(SDL_Renderer *d, const SDL_Rect *rect) {
     if (d->drawColor.a == 255 || d->blend == SDL_BLENDMODE_NONE) {
         if (rect->w == 1 && rect->h >= 3) {
             fillRectChar_Canvas(canvas, *(const iRect *) rect, transparent_VisChar,
-                                0x2503 /* box drawings heavy vertical */, drawColor_(d));   
+                                0x2503 /* box drawings heavy vertical */, drawColor_(d));
         }
         else if (rect->h == 0 && rect->w > 1) {
             fillRectChar_Canvas(canvas, init_Rect(rect->x, rect->y, rect->w, 1),
@@ -472,7 +501,8 @@ void SDL_RenderFillRect(SDL_Renderer *d, const SDL_Rect *rect) {
                                 drawColor_(d));
         }
         else {
-            fillRect_Canvas(canvas, *(const iRect *) rect, d->drawColorId);
+            fillRectChar_Canvas(canvas, init_Rect(rect->x, rect->y, rect->w, rect->h),
+                                d->drawColorId, ' ', d->textColorId);
         }
     }
 }
@@ -538,11 +568,11 @@ static const struct { uint32_t ucp; const char *str; } simpleChars_[] = {
     { 0x0263b, ":)" },
     { 0x0264a, "][" }, /* gemini */
     { 0x026a0, "/\\" }, /* warning */
-    { 0x026f2, "`m'" }, /* fountain */   
+    { 0x026f2, "`m'" }, /* fountain */
     { 0x02714, "X" }, /* check */
     { 0x02728, "`*`" }, /* sparkles */
     { 0x02795, "+" },
-    { 0x027a4, ">>" }, /* right arrowhead */    
+    { 0x027a4, ">>" }, /* right arrowhead */
     { 0x027e9, ">" }, /* closed folder */
     { 0x02913, "_v_" }, /* arrow down to bar */
     { 0x02937, "\u2514\u2500" }, /* arrow pointing downwards then curving rightwards */
@@ -553,7 +583,7 @@ static const struct { uint32_t ucp; const char *str; } simpleChars_[] = {
     { 0x1f31f, ":*:" }, /* glowing star */
     { 0x1f337, "\\U/" }, /* tulip */
     { 0x1f33b, "O)-" }, /* sunflower */
-    { 0x1f392, "c3" }, /* backpack */    
+    { 0x1f392, "c3" }, /* backpack */
     { 0x1f3e0, "<#" }, /* home */
     { 0x1f3e1, "<#:" }, /* house with garden */
     { 0x1f3f7, " /*" }, /* label tag */
@@ -596,7 +626,7 @@ static const struct { uint32_t ucp; const char *str; } simpleChars_[] = {
     { 0x1f607, "O:)" },
     { 0x1f608, "}:)" },
     { 0x1f609, ";-)" },
-    { 0x1f60a, ":-D" },
+    { 0x1f60a, ":-)" },
     { 0x1f60b, ":)_" },
     { 0x1f60c, "3-)" },
     { 0x1f60d, "<3" },
@@ -683,12 +713,12 @@ static const char *simpleCharStr_(uint32_t ucp) {
                     break; /* too short */
                 }
                 if (ucp < midUcp) {
-                    range.end = mid;                    
+                    range.end = mid;
                 }
                 else {
                     range.start = mid;
                 }
-            }           
+            }
         }
         if (ucp >= 0x2500 && ucp <= 0x257f) {
             return NULL; /* box drawing usually supported */
@@ -709,7 +739,7 @@ iLocalDef void renderChar_(SDL_Renderer *d, iVisChar *vis, uint32_t ucp) {
     if (d->textFillColorId != transparent_VisChar) {
         vis->bg = d->textFillColorId;
     }
-    vis->fg = d->textColorId;                    
+    vis->fg = d->textColorId;
 }
 
 void SDL_RenderDrawUnicode(SDL_Renderer *d, int x, int y, uint32_t ucp) {
@@ -732,6 +762,11 @@ void SDL_RenderDrawUnicode(SDL_Renderer *d, int x, int y, uint32_t ucp) {
             }
         }
         else {
+            if (ucp == 0x2014) {
+                /* Em dashes are too wide for text mode so use En dash instead. This is more
+                   compatible with fonts and terminal emulators. */
+                ucp--;
+            }
             renderChar_(d, at_Canvas(canvas, pos), ucp);
         }
     }
@@ -745,5 +780,8 @@ int SDL_UnicodeWidth(SDL_Renderer *d, uint32_t ucp) {
     if (simple) {
         return strlen(simple);
     }
+    if (ucp == 0x2014) {
+        return 1; /* We're replacing Em with an En dash */
+    }
     return width_Char(ucp);
 }
